#include "all_logcpp_libraries.hpp"

#include "fast_library.hpp"

// Get log4cpp logger from main program
extern log4cpp::Category& logger;

// Global configuration map
extern std::map<std::string, std::string> configuration_map;
